#!/usr/bin/env python3

# standards
from typing import Dict, Type

# hublot
from .base import Engine


ALL_ENGINES: Dict[str, Type[Engine]] = {}


def register_engine(engine_class: Type[Engine]) -> None:
    ALL_ENGINES[engine_class.id] = engine_class



#!/usr/bin/env python3

# standards
from dataclasses import replace
from pathlib import Path
from typing import Iterable, Tuple

# 3rd parties
import pytest

# hublot
from hublot import HttpClient, Response
from hublot.cache.storage import DiskStorage
from hublot.logs import LogEntry
from .utils import dummy_compiled_request, dummy_response


def iter_pairs(client: HttpClient) -> Iterable[Tuple[dict, Response]]:
    for url in ("http://one/", "http://two/"):
        for method, body in [("GET", None), ("POST", b"dummy body")]:
            for headers in ({}, {"X-Test": "1"}, {"X-Test": "2"}):
                # NB there's more comprehensive tests for cache key equivalent in `test_cache_keys.py`, but here's a sample
                request_kwargs = {"url": url, "method": method, "headers": headers, "data": body}
                creq = dummy_compiled_request(client, **request_kwargs)
                res = dummy_response(creq, from_cache=True)
                yield request_kwargs, res


def test_cache(reinstantiable_client) -> None:
    client = reinstantiable_client()
    pairs = [(dummy_compiled_request(client, **req), res) for req, res in iter_pairs(client)]
    log_entries = [LogEntry(creq) for creq, _res_unused in pairs]
    cache = client.cache
    for (creq, _res_unused), log in zip(pairs, log_entries):
        assert cache.get(creq, log) is None  # else test is invalid
        assert log.cached is False
    for (creq, res), log in zip(pairs, log_entries):
        cache.put(creq, log, res)
    cache = reinstantiable_client().cache
    for (creq, res), log in zip(pairs, log_entries):
        assert cache.get(creq, log) == res
        assert log.cached is True


def test_client_caching(mocker, reinstantiable_client) -> None:
    client = reinstantiable_client()
    for req, res in iter_pairs(client):
        res = replace(res, from_cache=False)
        request = mocker.patch.object(client.engine, "request", return_value=res)
        assert client.fetch(**req) == res
        request.assert_called_once()
    client = reinstantiable_client()
    for req, res in iter_pairs(client):
        res = replace(res, from_cache=True)
        request = mocker.patch.object(client.engine, "request", return_value=res)
        assert client.fetch(**req) == res
        request.assert_not_called()


def test_http_errors_are_cached(client, server) -> None:
    one = client.get(f"{server}/fail-with-random-value", raise_for_status=False)
    assert one.status_code == 500
    two = client.get(f"{server}/fail-with-random-value", raise_for_status=False)
    assert two.status_code == 500
    three = client.get(f"{server}/fail-with-random-value", raise_for_status=False, force_cache_stale=True)
    assert three.status_code == 500
    assert one.text == two.text
    assert two.text != three.text


def test_repeated_http_headers_are_cached(reinstantiable_client, server) -> None:
    client = reinstantiable_client()
    res = client.get(f"{server}/cookies/set-two-cookies")
    assert res.headers.get("Set-Cookie") == "a=1; b=2"
    assert sorted(f"{c.name}={c.value!r}" for c in client.cookies) == ["a='1'", "b='2'"]
    unique = res.text

    client = reinstantiable_client()
    res = client.get(f"{server}/cookies/set-two-cookies")
    assert res.text == unique  # it was cached
    assert res.headers.get("Set-Cookie") == "a=1; b=2"
    assert res.headers.get_all("Set-Cookie") == ["a=1", "b=2"]

    assert sorted(f"{c.name}={c.value!r}" for c in client.cookies) == ["a='1'", "b='2'"]
    assert [f"{c.name}={c.value!r}" for c in client.cookies] == ["a='1'", "b='2'"]


def test_cache_wont_save_body_with_wrong_length(client) -> None:
    creq = dummy_compiled_request(
        client,
        data=b"More than one byte",
        headers={"Content-Length": "1"},
    )
    response = dummy_response(creq)
    with pytest.raises(Exception) as ex:
        client.cache.put(creq, LogEntry(creq), response)
    assert "body has 18 bytes but Content-Length is 1" in str(ex)


def test_cache_wont_save_get_request_with_content_length(client) -> None:
    creq = dummy_compiled_request(
        client,
        method="GET",
    )
    creq.headers["Content-Length"] = "999"
    response = dummy_response(creq)
    with pytest.raises(Exception) as ex:
        client.cache.put(creq, LogEntry(creq), response)
    assert "body has 0 bytes but Content-Length is 999" in str(ex)


def test_cache_can_handle_empty_post_request(client, server) -> None:
    for _from_cache_unused in (False, True):
        res = client.fetch(f"{server}/echo", data={})
        obtained = res.json()
        obtained.pop("headers")
        assert obtained == {"method": "POST", "args": {}, "data": ""}


def test_cache_can_be_specified_as_path() -> None:
    client = HttpClient(cache=Path("/cache"))
    assert client.cache is not None
    assert isinstance(client.cache.storage, DiskStorage)
    assert client.cache.storage.root_path == Path("/cache")


def test_cache_cannot_be_specified_as_str() -> None:
    # The idea of not accepting strings is mostly to just keep the client code readable, and maybe preserve forward flexibility.
    # But if this ever a problem, we could consider auto-converting str objects to Path objects.
    with pytest.raises(TypeError):
        HttpClient(cache="/cache")


def test_from_cache_attribute(client, server) -> None:
    for from_cache in (False, True):
        res = client.fetch(f"{server}/hello")
        assert res.from_cache == from_cache


@pytest.mark.skip("Flask 2 no longer lets us send a response without a reason")
def test_cache_can_handle_empty_http_reason(client, server) -> None:  # pragma: no cover
    for _from_cache_unused in (False, True):
        res = client.fetch(f"{server}/no-reason", raise_for_status=False)
        assert res.reason == ""
        assert res.text == "hello"



#!/usr/bin/env python3

# 3rd parties
import pytest

# hublot
from hublot import ConnectionError, HttpClient, Request


@pytest.mark.parametrize(
    ["request_kwargs", "expected_payload"],
    [
        pytest.param(
            dict(params={"a": "b"}),
            {
                "method": "GET",
                "headers": {},
                "args": {"a": "b"},
                "data": "",
            },
            id="simple get",
        ),
        pytest.param(
            dict(data={"a": "b"}),
            {
                "method": "POST",
                "headers": {"Content-Length": "3", "Content-Type": "application/x-www-form-urlencoded"},
                "args": {},
                "data": "a=b",
            },
            id="simple post",
        ),
        pytest.param(
            dict(json={"a": "b"}),
            {
                "method": "POST",
                "headers": {"Content-Length": "9", "Content-Type": "application/json"},
                "args": {},
                "data": '{"a":"b"}',
            },
            id="simple JSON post",
        ),
        pytest.param(
            dict(method="SLURP"),
            {
                "method": "SLURP",
                "headers": {"Content-Length": "0"},
                "args": {},
                "data": "",
            },
            id="custom method, no body",
        ),
        pytest.param(
            dict(method="POST", data=b"\x00\x01\x02"),
            {
                "method": "POST",
                "headers": {"Content-Length": "3", "Content-Type": "application/x-www-form-urlencoded"},
                "args": {},
                "data": r"\x00\x01\x02",
            },
            id="post unprintable bytes",
        ),
        pytest.param(
            dict(method="POST", data=b"\x00\x01\x02", headers={"Content-Type": "application/whatever"}),
            {
                "method": "POST",
                "headers": {"Content-Length": "3", "Content-Type": "application/whatever"},
                "args": {},
                "data": r"\x00\x01\x02",
            },
            id="post with custom content-type",
        ),
    ],
)
def test_http_requests(engines, server, request_kwargs, expected_payload) -> None:
    client = HttpClient(engines=engines)
    res = client.fetch(Request(f"{server}/echo", **request_kwargs))
    payload = res.json()
    print(payload)
    assert isinstance(payload, dict)
    for key in ["Accept", "Accept-Encoding", "Host", "User-Agent"]:
        payload["headers"].pop(key, None)
    assert payload == expected_payload


def test_long_response(engines, server) -> None:
    client = HttpClient(engines=engines)
    length = 12 * 1024 * 1024
    res = client.fetch(f"{server}/bytes", params={"length": length})
    assert res.content == b"\x00" * length


@pytest.mark.skip("Need proxy servers to test against (both HTTP and HTTPS, ideally)")  # pragma: no cover
def test_proxies(engines, server) -> None:
    client = HttpClient(engines=engines)
    res = client.fetch("http://hublot.test/", proxies={"http": server})
    assert res.content == ""


def test_unknown_hostname(engines) -> None:
    client = HttpClient(engines=engines)
    with pytest.raises(ConnectionError):
        client.get("http://thishostdoesnotexist.test/")



#!/usr/bin/env python3

# standards
from io import BytesIO
import logging
import re
from typing import Any, Optional
from urllib.parse import urlparse

# 3rd parties
try:
    import pycurl

    HAVE_PYCURL = True
except ImportError:
    HAVE_PYCURL = False

# hublot
from ..config import Config
from ..datastructures import CompiledRequest, ConnectionError, Headers, HublotException, Response
from .base import Engine
from .register import register_engine


LOGGER = logging.getLogger(__name__)


RE_STATUS = re.compile(
    r"^\s* HTTP/\d+(?:\.\d+)? \s+ (\d\d\d) \s* (?: (\S.+?) \s* )? $",
    flags=re.M | re.X,
)

RE_HEADER = re.compile(
    r"^\s* ([^:]+?) \s*:\s* (.+?) \s*$",
    flags=re.M | re.X,
)


class PyCurlEngine(Engine):

    id = "pycurl"

    def __init__(self) -> None:
        if not HAVE_PYCURL:
            raise ImportError("pycurl is not installed; maybe try 'pip install hublot[pycurl]'")
        self.curl: Any = pycurl.Curl()  # pylint: disable=c-extension-no-member

    def short_code(self) -> str:
        return "pc"

    def request(self, creq: CompiledRequest, config: Config) -> Response:
        c = self.curl

        c.setopt(c.URL, creq.url)
        c.setopt(c.CUSTOMREQUEST, creq.method)
        c.setopt(
            c.HTTPHEADER,
            [f"{key}: {value}".encode("ISO-8859-1") for key, value in creq.headers.items()],
        )
        if creq.data is not None:
            c.setopt(c.POSTFIELDS, creq.data)

        if config.proxies:
            scheme = urlparse(creq.url).scheme
            proxy = config.proxies.get(scheme)
            if proxy:
                c.setopt(c.PROXY, proxy)

        c.setopt(c.ACCEPT_ENCODING, "")  # accept all curl-supported encodings
        c.setopt(c.TIMEOUT, config.timeout)
        c.setopt(c.SSL_VERIFYHOST, 2 if config.verify else 0)
        c.setopt(c.SSL_VERIFYPEER, 1 if config.verify else 0)

        headers = Headers()
        status_code: list[int] = []
        reason: list[Optional[str]] = []

        c.setopt(c.HEADERFUNCTION, lambda line: self.handle_header_line(headers, status_code, reason, line))
        output_bytes = BytesIO()
        c.setopt(c.WRITEFUNCTION, output_bytes.write)
        try:
            c.perform()
        except pycurl.error as error:  # pylint: disable=c-extension-no-member
            error_code, message = error.args
            if error_code == 6:
                raise ConnectionError(message) from error
            else:
                raise HublotException() from error

        return Response(
            request=creq,
            from_cache=False,
            history=[],
            status_code=status_code[0],
            reason=reason[0] if reason else None,
            headers=headers,
            content=output_bytes.getvalue(),
        )

    @staticmethod
    def handle_header_line(
        headers: Headers,
        status_code: list[int],
        reason: list[Optional[str]],
        line_bytes: bytes,
    ) -> None:
        # HTTP standard specifies that headers are encoded in iso-8859-1
        line = line_bytes.decode("ISO-8859-1")
        if not line.strip():
            return
        if not status_code:
            match = RE_STATUS.match(line)
            if match:
                status_code.append(int(match[1]))
                reason.append(match[2])
            else:  # pragma: no cover
                LOGGER.warning("Malformed status line: %r", line)
        else:
            match = RE_HEADER.match(line)
            if match:
                headers.add(match[1], match[2])
            else:  # pragma: no cover
                LOGGER.warning("Malformed header line: %r", line)


register_engine(PyCurlEngine)



#!/usr/bin/env python3

# 3rd parties
import pytest

# hublot
from hublot import Headers


def test_empty_headers() -> None:
    headers = Headers()
    assert not headers
    assert headers.get("anything") is None
    with pytest.raises(KeyError):
        print(headers["anything"])
    assert list(headers.get_all("Anything")) == []
    assert list(headers) == []
    assert list(headers.keys()) == []
    assert list(headers.items()) == []
    assert headers == Headers()
    assert headers != Headers({"My-Key": "My-Value"})
    assert headers != {}  # never equal to sth that isn't a Headers object


def test_add_method() -> None:
    headers = Headers()
    headers.add("My-Key", "My-Value")
    assert headers
    assert "My-Key" in headers
    assert headers["My-Key"] == "My-Value"
    with pytest.raises(KeyError):
        print(headers["anything else"])
    assert list(headers.get_all("My-Key")) == ["My-Value"]
    assert list(headers) == ["My-Key"]
    assert list(headers.keys()) == ["My-Key"]
    assert list(headers.items()) == [("My-Key", "My-Value")]
    assert headers == Headers({"My-Key": "My-Value"})
    assert headers != {"My-Key": "My-Value"}


def test_constructor_arg() -> None:
    headers = Headers({})
    headers.add("My-Key", "My-Value")
    assert headers == Headers({"My-Key": "My-Value"})


def test_case_insensitivity() -> None:
    headers = Headers()
    headers.add("My-Key", "My-Value")
    assert headers["My-Key"] == "My-Value"
    assert headers["my-key"] == "My-Value"
    assert headers["mY-kEY"] == "My-Value"
    assert "My-Key" in headers
    assert "my-key" in headers
    assert "mY-kEY" in headers
    assert list(headers.get_all("my-key")) == ["My-Value"]


def test_key_case_is_preserved() -> None:
    headers = Headers()
    headers.add("my-key", "My-Value")
    assert list(headers) == ["my-key"]
    assert list(headers.keys()) == ["my-key"]
    assert list(headers.items()) == [("my-key", "My-Value")]
    assert list(headers.items(normalise_keys=True)) == [("My-Key", "My-Value")]


def test_eq() -> None:
    assert Headers({"My-Key": "My-Value"}) == Headers({"my-key": "My-Value"})
    assert Headers({"My-Key": "My-Value"}) != Headers({"My-Key": "vALUE"})


def test_multiple_values() -> None:
    headers = Headers()
    headers.add("My-Key", "My-Value-1")
    headers.add("My-Key", "My-Value-2")
    headers.add("My-Key", "My-Value-3")
    assert list(headers) == ["My-Key"]
    assert list(headers.items()) == [("My-Key", "My-Value-1"), ("My-Key", "My-Value-2"), ("My-Key", "My-Value-3")]
    assert headers.get("My-Key") == "My-Value-1; My-Value-2; My-Value-3"


def test_setdefault() -> None:
    headers = Headers()
    headers.setdefault("My-Key", "My-Value-1")
    headers.setdefault("My-Key", "My-Value-2")
    assert headers["My-Key"] == "My-Value-1"


def test_repr() -> None:
    # We don't check the specifics of the repr format, but we do check that the values are included
    headers = Headers()
    headers.add("My-Key", "My-Value-1")
    headers.add("My-Key", "My-Value-2")
    assert "My-Key" in repr(headers)
    assert "My-Value-1" in repr(headers)
    assert "My-Value-2" in repr(headers)



#!/usr/bin/env python3

# hublot
from hublot import HttpClient


def test_cookies_are_saved(client, server):
    assert client.get(f"{server}/cookies/get").json() == {}
    client.get(f"{server}/cookies/set?coo=kie")
    assert client.get(f"{server}/cookies/get").json() == {"coo": "kie"}


def test_cookies_are_available_via_client(cache, server):
    for _ in (1, 2):
        client = HttpClient(cache=cache)  # keep cache, clear cookies
        client.get(f"{server}/cookies/set?coo=kie")
        assert dict(client.cookies) == {"coo": "kie"}


def test_cookies_are_available_via_response(cache, server):
    for _ in (1, 2):
        client = HttpClient(cache=cache)  # keep cache, clear cookies
        response = client.get(f"{server}/cookies/set?coo=kie")
        assert dict(response.cookies) == {"coo": "kie"}


def test_cookies_are_set_when_running_from_cache(reinstantiable_client, server):
    for _ in (1, 2):
        client = reinstantiable_client()
        assert client.get(f"{server}/cookies/get").json() == {}
        client.get(f"{server}/cookies/set?coo=kie")
        assert client.get(f"{server}/cookies/get").json() == {"coo": "kie"}


def test_cached_redirects(cache, server):
    for _ in (1, 2):
        client = HttpClient(cache=cache)
        client.get(f"{server}/redirect/chain/1")
        cookies = {c.name: c.value for c in client.cookies}
        assert cookies == {"redirect1": "yes", "redirect2": "yes", "redirect3": "yes"}


def test_cookies_can_be_disabled(cache, server):
    for _ in (1, 2):
        client = HttpClient(cache=cache, cookies_enabled=False)
        assert client.get(f"{server}/cookies/get").json() == {}
        client.get(f"{server}/cookies/set?coo=kie")
        assert client.get(f"{server}/cookies/get").json() == {}


def test_multiple_cookies_headers(reinstantiable_client, server):
    for reading_from_cache in (False, True):
        client = reinstantiable_client()
        response = client.get(f"{server}/cookies/set-two-cookies")
        if not reading_from_cache:
            # Check that the mock server returns cookies in 2 separate headers, else this test isn't testing what it meant to test.
            # Can't run this test on the second iteration, because we don't have `response.raw` when coming from cache.
            response_cookies = sorted((key, value) for key, value in response.headers.items() if key == "Set-Cookie")
            assert response_cookies == [("Set-Cookie", "a=1"), ("Set-Cookie", "b=2")]
        assert sorted(response.cookies.items()) == [("a", "1"), ("b", "2")]
        assert sorted(client.cookies.items()) == [("a", "1"), ("b", "2")]


def test_response_that_redirects_to_same_url_with_cookie(client, server):
    assert client.get(f"{server}/self-redirect").text == "ok"


def test_set_cookie_manually(client, server):
    client.cookies.set("x", "1")
    assert client.get(f"{server}/cookies/get").json() == {"x": "1"}


def test_cookies_can_be_cleared(client, server):
    assert client.get(f"{server}/cookies/get").json() == {}
    client.get(f"{server}/cookies/set?coo=kie")
    assert client.get(f"{server}/cookies/get").json() == {"coo": "kie"}
    client.cookies.clear()
    assert client.get(f"{server}/cookies/get").json() == {}



#!/usr/bin/env python3

# standards
from contextlib import contextmanager
from datetime import timedelta
from time import sleep, time
from typing import Dict, Iterator, Optional, Sequence
from urllib.parse import urljoin, urlparse

# 3rd parties
#
# Even when using other engines, we rely on requests's utilities for handling cookies here, otherwise we'd have several wheels to
# reinvent
from requests.cookies import RequestsCookieJar

# hublot
from .cache import CacheKey, CacheSpec, UserSpecifiedCacheKey, load_cache
from .compile import compile_request
from .config import Config
from .datastructures import CompiledRequest, Request, Requestable, Response, TooManyRedirects, get_cookies_from_response
from .decorator import SCRAPER_LOCAL
from .engines import EngineSpec, load_engine_pool
from .logs import LOGGER, LogEntry


class HttpClient:
    """
    Core class for this package. Meant as a mostly-drop-in replacement for `requests.Session`, but handles caching, courtesy
    sleep, and can use multiple engines for performing the actual HTTP transactions.
    """

    def __init__(
        self,
        cache: CacheSpec = None,
        engines: Sequence[EngineSpec] = ("requests",),
        **config_kwargs,
    ) -> None:
        self.config = Config(**config_kwargs)
        self.cache = load_cache(cache, self.config)
        self.engine = load_engine_pool(engines)
        self.cookies = RequestsCookieJar()
        self.last_request_per_host: Dict[str, float] = {}

    def fetch(
        self,
        url: Requestable,
        cache_key: Optional[UserSpecifiedCacheKey] = None,
        **kwargs,
    ) -> Response:
        """
        Main public method for this class.

        NB the first parameter is called `url`, even though it could also be a `Request` object, simply because that's shorter
        and nicer than calling it `requestable`. Hey, `urllib.request.urlopen` does the same thing, so.
        """
        config, request_kwargs = self.config.derive_using_kwargs(**kwargs)
        req = Request.build(url, **request_kwargs)
        history: list[Response] = []
        for redirect_count in range(config.max_redirects):
            res = self._fetch_without_redirect(
                req,
                cache_key,
                config,
                is_redirect=(redirect_count > 0),
            )
            if config.allow_redirects and res.is_redirect:
                req = req.replace(
                    url=urljoin(res.url, res.headers["Location"]),
                    params={},
                )
                if res.status_code not in (307, 308):
                    req = req.replace(method="GET", data=None, json=None)
                if cache_key:
                    cache_key = CacheKey.parse(cache_key).next_in_sequence()
                history.append(res)
            else:
                if config.raise_for_status:
                    res.raise_for_status()
                res.history = history
                return res
        raise TooManyRedirects(f"Exceeded {config.max_redirects} redirects")

    def _fetch_without_redirect(
        self,
        req: Request,
        cache_key: Optional[UserSpecifiedCacheKey],
        config: Config,
        is_redirect: bool,
    ) -> Response:
        frame = SCRAPER_LOCAL.stack[-1]
        if frame.num_retries > 0:
            config.force_cache_stale = True
            config.courtesy_sleep = timedelta(0)
        creq = compile_request(req, config, self.cookies, frame.num_retries)
        log = LogEntry(creq, is_redirect)
        res = self._read_response(cache_key, creq, config, is_redirect, log)
        if config.cookies_enabled:
            get_cookies_from_response(self.cookies, res)
        LOGGER.info("%s", log)
        return res

    def _read_response(
        self,
        cache_key: Optional[UserSpecifiedCacheKey],
        creq: CompiledRequest,
        config: Config,
        is_redirect: bool,
        log: LogEntry,
    ) -> Response:
        """
        Either read the Response from cache, or perform the HTTP transaction and save the response to cache
        """
        if self.cache and not config.force_cache_stale:
            res = self.cache.get(creq, log, config, cache_key)
            if res is not None:
                res.from_cache = True
                return res
        with self._sleep_if_needed(config, creq, is_redirect, log):
            log.engine_short_code = self.engine.short_code()
            res = self.engine.request(creq, config)
        assert not res.from_cache  # would be an engine bug
        if self.cache:
            self.cache.put(creq, log, res, config, cache_key)
        return res

    @contextmanager
    def _sleep_if_needed(
        self,
        config: Config,
        creq: CompiledRequest,
        is_redirect: bool,
        log: LogEntry,
    ) -> Iterator[None]:
        host = urlparse(creq.url).hostname or ""
        last_request = self.last_request_per_host.get(host, 0)
        if config.courtesy_sleep and not is_redirect:
            delay_seconds = (last_request + config.courtesy_sleep.total_seconds()) - time()
            if delay_seconds > 0:
                log.courtesy_seconds = delay_seconds
                sleep(delay_seconds)
        try:
            yield
        finally:
            # NB we store the time after the request is complete
            self.last_request_per_host[host] = time()

    ### for a thin veneer of Requests compatibility

    def request(self, method: str, url: str, **kwargs) -> Response:
        return self.fetch(url, method=method, **kwargs)

    def get(self, url: str, **kwargs) -> Response:
        return self.fetch(url, method="GET", **kwargs)

    def post(self, url: str, data=None, **kwargs) -> Response:
        return self.fetch(url, method="POST", data=data, **kwargs)



#!/usr/bin/env python3

from .cache import Cache, CacheKey
from .client import HttpClient
from .datastructures import (
    CharsetDetectionFailure,
    ConnectionError,
    HttpError,
    Headers,
    HublotException,
    Request,
    Requestable,
    Response,
    TooManyRedirects,
)
from .decorator import retry_on_scraper_error
from .engines import register_engine
from .logs import LOGGER as logger, basic_logging_config

__all__ = [
    "Cache",
    "CacheKey",
    "CharsetDetectionFailure",
    "ConnectionError",
    "Headers",
    "HttpClient",
    "HttpError",
    "HublotException",
    "Request",
    "Requestable",
    "Response",
    "TooManyRedirects",
    "basic_logging_config",
    "logger",
    "register_engine",
    "retry_on_scraper_error",
]



#!/usr/bin/env python3

# standards
from typing import Optional, Sequence
from uuid import UUID

# hublot
from ..config import Config
from ..datastructures import CompiledRequest, Response
from ..decorator import SCRAPER_LOCAL
from .base import Engine


class EnginePool(Engine):

    id = "engine-pool"

    def __init__(self, engines: Sequence[Engine]):
        self.engines = tuple(engines)
        self.rotation = 0
        self.last_state: tuple[Optional[UUID], int] = (None, 0)

    def _get_next_engine(self, save_state: bool = False) -> Engine:
        frame = SCRAPER_LOCAL.stack[-1]
        if frame.num_retries == 0:
            # This is our first attempt at this request, use which ever engine is at the front of the queue, and don't rotate. As
            # long as that engine works, we'll keep using it
            return self.engines[self.rotation]
        else:
            state = (frame.request_uuid, frame.num_retries)
            if self.last_state == state:
                # We've been called already for this attempt at performing this request. This happens because both `short_code` and
                # `request` call this function. Don't rotate, just return the same engine that we've already returned for this
                # attempt at performing this request.
                return self.engines[self.rotation]
            else:
                # We failed in our previous attempt at performing this request, and now we're attempting again. Rotate the engines.
                rotation = (self.rotation + 1) % len(self.engines)
                if save_state:
                    self.rotation = rotation
                    self.last_state = state
                return self.engines[rotation]

    def short_code(self) -> str:
        return self._get_next_engine().short_code()

    def request(self, creq: CompiledRequest, config: Config) -> Response:
        return self._get_next_engine(save_state=True).request(creq, config)



#!/usr/bin/env python3

# standards
from typing import Sequence, Union

# hublot
from .base import Engine
from .curlcmd import CurlCmdEngine
from .pool import EnginePool
from .pycurl import PyCurlEngine
from .register import ALL_ENGINES, register_engine
from .requests import RequestsEngine


__all__ = [
    "ALL_ENGINES",
    "CurlCmdEngine",
    "Engine",
    "EnginePool",
    "PyCurlEngine",
    "RequestsEngine",
    "load_engine_pool",
    "register_engine",
]


EngineSpec = Union[Engine, str]


def load_engine_pool(engine_specs: Sequence[EngineSpec]) -> EnginePool:
    return EnginePool(
        engines=list(map(_get_engine_instance, engine_specs)),
    )


def _get_engine_instance(spec: EngineSpec) -> Engine:
    if isinstance(spec, Engine):
        return spec
    if ":" in spec:
        # You can pass string args to the Engine constructor by putting then after a colon. At the moment only the CurlCmdEngine
        # uses that. If an engine needs more complex constructor args, the client code can just instantiate the engine class
        # itself. This is just a minor convenience to avoid the client having to import the CurlCmdEngine class.
        engine_id, *engine_args = spec.split(":")
    else:
        engine_id = spec
        engine_args = []
    engine_class = ALL_ENGINES[engine_id]
    return engine_class(*engine_args)



#!/usr/bin/env python3

# standards
import re

# hublot
from hublot import retry_on_scraper_error


def test_basic_logging(client, server, captured_logs):
    engine = client.engine.short_code()
    client.get(f"{server}/hello")
    assert re.search(rf"^\[\w\w\w/\w{{13}}\] \[{engine}\]     {server}/hello\n$", captured_logs())
    client.get(f"{server}/hello")
    assert re.search(rf"^\[\w\w\w/\w{{13}}\] \[cached\] {server}/hello\n$", captured_logs())


def test_logging_courtesy_sleep(client, server, captured_logs):
    engine = client.engine.short_code()
    client.get(f"{server}/echo?x=1")
    assert re.search(rf"^\[\w\w\w/\w{{13}}\] \[{engine}\]     {server}/echo\?x=1\n$", captured_logs())
    client.get(f"{server}/echo?x=2")
    assert re.search(rf"^\[\w\w\w/\w{{13}}\] \[{engine}\+5s\]  {server}/echo\?x=2\n$", captured_logs())


def test_logging_redirects(client, server, captured_logs):
    engine = client.engine.short_code()
    client.get(f"{server}/redirect/chain/1")
    assert re.search(
        r"^"
        rf"\[\w\w\w/\w{{13}}\] \[{engine}\]     {server}/redirect/chain/1\n"
        rf"\[\w\w\w/\w{{13}}\] \[{engine}\]     -> {server}/redirect/chain/2\n"
        rf"\[\w\w\w/\w{{13}}\] \[{engine}\]     -> {server}/redirect/chain/3\n"
        r"$",
        captured_logs(),
    )


def test_logging_on_retry(client, server, unique_key, captured_logs):
    engine = client.engine.short_code()

    @retry_on_scraper_error
    def scrape():
        client.get(f"{server}/fail-twice-then-succeed/{unique_key}")
        return "ok"

    scrape()
    assert re.search(
        r"^"
        rf"\[\w\w\w/\w{{13}}\] \[{engine}\]     {server}/fail-twice-then-succeed/{unique_key}\n"
        "HttpError: 500 .+ sleeping 1s\n"
        rf"\[\w\w\w/\w{{13}}\] \[{engine}\]     {server}/fail-twice-then-succeed/{unique_key}\n"
        "HttpError: 500 .+ sleeping 5s\n"
        rf"\[\w\w\w/\w{{13}}\] \[{engine}\]     {server}/fail-twice-then-succeed/{unique_key}\n"
        r"$",
        captured_logs(),
    )



#!/usr/bin/env python3

# standards
from pathlib import Path
from typing import Optional, Union

# hublot
from ..config import Config
from ..datastructures import CompiledRequest, Response
from ..logs import LogEntry
from .key import CacheKey, UserSpecifiedCacheKey
from .storage import DiskStorage, Storage


class Cache:

    def __init__(
        self,
        storage: Storage,
        config: Config = Config(),
    ) -> None:
        self.storage = storage
        self.config = config
        self.max_age_overall = config.max_cache_age
        self.has_been_pruned = False

    def get(
        self,
        creq: CompiledRequest,
        log: LogEntry,
        config: Optional[Config] = None,
        key: Optional[UserSpecifiedCacheKey] = None,
    ) -> Optional[Response]:
        """
        Looks up the given `CompiledRequest`, and returns the corresponding `Response` if it was in cache, or `None` otherwise.
        """
        if config is None:
            config = self.config
        key = CacheKey.parse(key) if key else CacheKey.compute(creq, config)
        if self.max_age_overall is not None and not self.has_been_pruned:
            self.storage.prune(self.max_age_overall)
            self.has_been_pruned = True
        max_age = config.max_cache_age
        if max_age is None or (self.max_age_overall is not None and max_age > self.max_age_overall):
            # The `max_age` passed to the method can't be greater than that given to the constructor, that would be inconsistent
            # with the cache pruning that happens and which is always based on the constructor-given `self.max_age_overall`
            max_age = self.max_age_overall
        res = self.storage.read(key, max_age)
        log.cached = res is not None
        log.cache_key_str = key.unique_str
        return res

    def put(
        self,
        creq: CompiledRequest,
        log: LogEntry,
        res: Response,
        config: Optional[Config] = None,
        key: Optional[UserSpecifiedCacheKey] = None,
    ) -> None:
        if config is None:
            config = self.config
        key = CacheKey.parse(key) if key else CacheKey.compute(creq, config)
        log.cache_key_str = key.unique_str
        self.storage.write(key, res)


CacheSpec = Union[Cache, Path, str, None]


def load_cache(
    cache: CacheSpec = None,
    config: Config = Config(),
) -> Optional["Cache"]:
    """
    Takes the `cache` param given to the `HttpClient` constructor, and returns a `Cache` instance, or `None`
    """
    if cache is None or isinstance(cache, Cache):
        return cache
    elif isinstance(cache, Path):
        return Cache(
            storage=DiskStorage(root_path=cache),
            config=config,
        )
    elif isinstance(cache, str) and cache.startswith("redis://"):  # pragma: no cover
        raise NotImplementedError  # some day
    else:
        raise TypeError(repr(cache))



#!/usr/bin/env python3

# standards
import re

# hublot
from hublot import HttpClient, retry_on_scraper_error
from hublot.config import Config
from hublot.datastructures import CompiledRequest, ConnectionError, Headers, Response
from hublot.engines import Engine, register_engine


class TestEngine(Engine):

    def short_code(self) -> str:
        return self.id[-1:]

    def request(self, creq: CompiledRequest, config: Config) -> Response:
        return Response(
            request=creq,
            from_cache=False,
            history=[],
            status_code=200,
            reason="OK",
            headers=Headers(),
            content=f"/{re.sub(r'.+/', '', creq.url)} from {self.id}".encode("UTF-8"),
        )


class Engine1(TestEngine):
    id = "engine1"

    def request(self, creq: CompiledRequest, config: Config) -> Response:
        if creq.url.endswith("/3"):
            raise ConnectionError("boom")  # network error
        return super().request(creq, config)


class Engine2(TestEngine):
    id = "engine2"

    def request(self, creq: CompiledRequest, config: Config) -> Response:
        if creq.url.endswith("/6"):
            return Response(
                request=creq,
                from_cache=False,
                history=[],
                status_code=404,  # HTTP status error
                reason="Not Found",
                headers=Headers(),
                content=b"Not Found",
            )
        return super().request(creq, config)


class Engine3(TestEngine):
    id = "engine3"

    def request(self, creq: CompiledRequest, config: Config) -> Response:
        if creq.url.endswith("/9"):
            return Response(
                request=creq,
                from_cache=False,
                history=[],
                status_code=200,
                reason="OK",
                headers=Headers(),
                content=b"Not the data you expected",
            )
        return super().request(creq, config)


def test_engine_pool_rotation_on_network_error() -> None:
    register_engine(Engine1)
    register_engine(Engine2)
    register_engine(Engine3)

    client = HttpClient(
        engines=[
            "engine1",
            "engine2",
            "engine3",
        ],
    )

    @retry_on_scraper_error
    def fetch(i) -> str:
        return parse(client.get(f"http://hublot.test/{i}").text)

    def parse(text: str) -> str:
        if " from " not in text:
            raise ValueError(text)
        return text

    results = list(map(fetch, range(12)))
    assert results == [
        "/0 from engine1",
        "/1 from engine1",
        "/2 from engine1",
        # engine1 gives a network error on /3, we're on to engine2
        "/3 from engine2",
        "/4 from engine2",
        "/5 from engine2",
        # engine2 returned an HTTP error on /6, we're on to engine3
        "/6 from engine3",
        "/7 from engine3",
        "/8 from engine3",
        # engine3 triggered a ValueError, we're back at engine1
        "/9 from engine1",
        "/10 from engine1",
        "/11 from engine1",
    ]



#!/usr/bin/env python3

# standards
from collections.abc import Generator, Iterator, Sized
from contextlib import contextmanager
from dataclasses import dataclass
from functools import wraps
import threading
from time import sleep
from typing import Callable, Optional, Sequence
from uuid import UUID, uuid4

# hublot
from .datastructures import HublotException
from .logs import LOGGER


@dataclass
class ThreadLocalStackFrame:
    request_uuid: Optional[UUID] = None
    num_retries: int = 0


class ThreadLocalStack(threading.local):

    def __init__(self) -> None:
        super().__init__()
        # each thread will magically get a different stack
        self.stack = [ThreadLocalStackFrame()]


SCRAPER_LOCAL = ThreadLocalStack()


def retry_on_scraper_error(
    no_parens_function: Optional[Callable] = None,
    *,
    error_types: Sequence[type] = (HublotException, ValueError),
    num_attempts: int = 5,
):
    if no_parens_function:
        # decorator is without parentheses
        return retry_on_scraper_error()(no_parens_function)

    @contextmanager
    def scraper_stack_frame():
        frame = ThreadLocalStackFrame()
        SCRAPER_LOCAL.stack.append(frame)
        try:
            yield frame
        finally:
            SCRAPER_LOCAL.stack.pop()

    error_types = tuple(error_types)

    def make_wrapper(function: Callable):
        @wraps(function)
        def wrapper(*args, **kwargs):
            with scraper_stack_frame() as frame:
                frame.request_uuid = uuid4()
                for attempt in range(num_attempts):
                    frame.num_retries = attempt
                    try:
                        payload = function(*args, **kwargs)
                        if isinstance(payload, (Iterator, Generator)) and not isinstance(payload, Sized):
                            # milk the generator so that all content is parsed
                            payload = list(payload)
                        return payload
                    except Exception as error:  # pylint: disable=broad-except
                        if isinstance(error, error_types) and attempt < num_attempts - 1:
                            delay = 5**attempt
                            LOGGER.error("%s: %s - sleeping %ds", type(error).__name__, error, delay)
                            sleep(delay)
                        else:
                            raise
            raise AssertionError("can't reach here")  # pragma: no cover

        return wrapper

    return make_wrapper



#!/usr/bin/env python3

# standards
import re

# hublot
from hublot import HttpClient


def test_default_user_agent(server):
    client = HttpClient()
    user_agent = client.get(f"{server}/echo").json()["headers"]["User-Agent"]
    assert re.search(r"^hublot/[\d\.]+$", user_agent), user_agent


def test_user_agent_constructor_kwarg(server):
    client = HttpClient(user_agent="Nyanya/10.8")
    user_agent = client.get(f"{server}/echo").json()["headers"]["User-Agent"]
    assert user_agent == "Nyanya/10.8"


def test_user_agent_method_kwarg(server):
    client = HttpClient(
        user_agent="this/gets/overriden",
    )
    payload = client.get(
        f"{server}/echo",
        user_agent="Prapra/12.12",
    ).json()
    user_agent = payload["headers"]["User-Agent"]
    assert user_agent == "Prapra/12.12"


def test_user_agent_in_headers(server):
    client = HttpClient()
    res = client.get(
        f"{server}/echo",
        headers={"User-Agent": "Bwabwa/7.3"},
    )
    user_agent = res.json()["headers"]["User-Agent"]
    assert user_agent == "Bwabwa/7.3"



#!/usr/bin/env python3

# standards
import json
from typing import Optional
from urllib.parse import urlencode

# 3rd parties
from requests.cookies import RequestsCookieJar, get_cookie_header

# hublot
from .config import Config
from .datastructures import CompiledRequest, Headers, Request


def compile_request(
    req: Request,
    config: Config,
    cookies: RequestsCookieJar,
    num_retries: int,
) -> CompiledRequest:
    headers = _compile_request_headers(config, req)
    data = _compile_request_data(req, headers)
    creq = CompiledRequest(
        url=_compile_request_url(req),
        method=req.method.upper() if req.method else ("POST" if data is not None else "GET"),
        headers=headers,
        data=data,
        num_retries=num_retries,
    )
    _add_cookies_to_request(cookies, creq)
    return creq


def _compile_request_headers(config: Config, req: Request) -> Headers:
    headers = Headers(req.headers)
    headers.setdefault("Accept", "*/*")
    if config.user_agent:
        headers.setdefault("User-Agent", config.user_agent)
    return headers


def _compile_request_url(req: Request) -> str:
    if not req.params:
        return req.url
    url = req.url.rstrip("?&")
    return (
        url
        + ("&" if "?" in url else "?")
        # We unconditionally use UTF-8 to encode URL params. In cases where this isn't desirable, the client should urlencode the
        # params itself, and pass them in str form, as part of the `url`.
        + urlencode(req.params, encoding="UTF-8")
    )


def _compile_request_data(req: Request, headers: Headers) -> Optional[bytes]:
    data = req.data
    if data is not None and req.json is not None:
        raise TypeError("Request cannot have both `data` and `json` set")
    if req.json is not None:
        headers.setdefault("Content-Type", "application/json")
        data = json.dumps(req.json, separators=(",", ":"))
    elif data is not None and not isinstance(data, (str, bytes)):
        headers.setdefault("Content-Type", "application/x-www-form-urlencoded")
        data = urlencode(data)
    if data is not None:
        if not isinstance(data, bytes):
            # We always encode as UTF-8. If another encoding is desired, the data should be pre-compiled to bytes by the caller
            data = data.encode("UTF-8")
        # We set a default content-type when uploading data. This is what Curl does, in fact I couldn't find a way of disabling
        # that.
        headers.setdefault("Content-Type", "application/x-www-form-urlencoded")
    if data is not None or (req.method and req.method.upper() not in ("HEAD", "GET")):
        # We set a Content-Length header even if `data` is None. This is what Requests does (in `prepare_content_length`), so for
        # consistency across engines we do the same.
        headers.setdefault("Content-Length", str(len(data or b"")))
    return data


def _add_cookies_to_request(cookies: RequestsCookieJar, creq: CompiledRequest) -> None:
    cookie = get_cookie_header(cookies, creq)
    if cookie is not None:
        creq.headers.add("Cookie", cookie)



#!/usr/bin/env python3

# standards
from threading import Thread
from time import sleep

# hublot
from hublot.decorator import SCRAPER_LOCAL, ThreadLocalStackFrame


def thread_body(value: int):
    num_sweeps = 50
    for _ in range(num_sweeps):
        SCRAPER_LOCAL.stack.append(ThreadLocalStackFrame(num_retries=value))
        sleep(0.01)
    for _ in range(num_sweeps):
        assert SCRAPER_LOCAL.stack.pop().num_retries is value
        sleep(0.01)


def test_thread_local_data():
    thread_a = Thread(target=thread_body, args=(True,))
    thread_b = Thread(target=thread_body, args=(False,))
    thread_a.start()
    thread_b.start()
    thread_a.join()
    thread_b.join()



#!/usr/bin/env python3

# standards
from abc import ABC, abstractmethod
from typing import ClassVar

# hublot
from ..config import Config
from ..datastructures import CompiledRequest, Response


class Engine(ABC):

    id: ClassVar[str]

    @abstractmethod
    def short_code(self) -> str:
        """
        Returns a two-letter code that will identify this engine in the logs
        """

    @abstractmethod
    def request(self, creq: CompiledRequest, config: Config) -> Response:
        """
        Perform one HTTP request, and return the response from the server, or raise an exception.

        Implementing subclasses should operate in a stateless fashion. In particular cookies will already be set as headers in the
        given `CompiledRequest`. The engine should not try to save or set cookies. This is important as it allows us to switch
        engines within a single session, carrying cookies across engines.

        Regardless of what `config.allow_redirects` it set to, the engine should not follow HTTP redirects. If a 30x response is
        received, the engine should just return that. Redirects are handled by the HttpClient class.

        The given CompiledRequest will already have Content-Length and Content-Type headers set, but we leave transport-related
        headers such as `Host`, `Connection` and `Accept-Encoding` headers up to the implementing subclass.
        """



#!/usr/bin/env python3

# standards
from dataclasses import dataclass
import logging
from typing import Iterator, List, Optional, Union

# hublot
from .datastructures import CompiledRequest

LOGGER = logging.getLogger("hublot")


@dataclass(frozen=False)
class LogEntry:
    creq: CompiledRequest
    is_redirect: bool = False
    cache_key_str: Optional[str] = None
    cached: Optional[bool] = None
    courtesy_seconds: Optional[float] = None
    engine_short_code: Optional[str] = None

    def _compose_line(self) -> Iterator[str]:
        if self.cache_key_str:
            yield f"[{self.cache_key_str}] "
        if self.cached:
            yield "[cached] "
        else:
            engine_and_sleep = self._compose_engine_and_sleep()
            if engine_and_sleep:
                yield f"{engine_and_sleep:8s} "
            else:
                yield "         "
        if self.is_redirect:
            yield "-> "
        yield self.creq.url
        if self.creq.data is not None:
            yield f" [{self.creq.method} {len(self.creq.data)} bytes]"

    def _compose_engine_and_sleep(self) -> Optional[str]:
        if self.cached:
            return "[cached]"
        parts: List[str] = []
        if self.engine_short_code:
            parts.append(self.engine_short_code)
        if self.courtesy_seconds and self.courtesy_seconds > 0.5:
            parts.append(f"{round(self.courtesy_seconds)}s")
        if not parts:
            return None
        return "[%s]" % "+".join(parts)

    def __str__(self) -> str:
        return "".join(self._compose_line())


def basic_logging_config(level: Union[int, str] = "INFO", propagate: bool = False) -> None:
    """
    Sets up logging for the common use case. Calls `logging.basicConfig`, lowers verbosity for the `urllib3` logger. If `propagate`
    is False (the default), a new handler will be attached to `hublot.LOGGER` that logs in a simple format to stderr, and does not
    propagate log events to the root logger.
    """
    if not isinstance(level, int):
        level = getattr(logging, level)
    logging.basicConfig(level=level)
    logging.getLogger("urllib3").setLevel(max(logging.WARNING, level))
    if not propagate and not LOGGER.handlers:
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(message)s", None, "%")
        handler.setFormatter(formatter)
        LOGGER.addHandler(handler)
        LOGGER.propagate = False



#!/usr/bin/env python3

from .cache import Cache, CacheSpec, load_cache
from .key import CacheKey, UserSpecifiedCacheKey
from .storage import DiskStorage, Storage

__all__ = [
    "Cache",
    "CacheKey",
    "CacheSpec",
    "DiskStorage",
    "Storage",
    "UserSpecifiedCacheKey",
    "load_cache",
]



#!/usr/bin/env python3

# standards
from collections import OrderedDict
from dataclasses import replace
from itertools import count, product

# 3rd parties
import pytest

# hublot
from hublot import Request
from hublot.cache import CacheKey
from hublot.config import Config
from hublot.logs import LogEntry
from .utils import dummy_compiled_request, dummy_response, iter_equal_pairs, iter_nonequal_pairs


def test_simple_cache_use(client):
    creq = dummy_compiled_request(client)
    log = LogEntry(creq)
    cache = client.cache
    assert cache.get(creq, log) is None
    assert log.cache_key_str is not None
    cache.put(creq, log, dummy_response(creq))
    assert cache.get(creq, log) == dummy_response(creq, from_cache=True)


EQUIVALENCIES = [
    # For every attribute of a CompiledRequest, lists groups values such that values within the same group should be cached under
    # the same key, but not across groups
    # Obviously requests using different methods should be cached separately
    [{"method": "GET"}],
    [{"method": "POST"}],
    # URLs are case sensitive
    [{"url": "http://cache-test/url"}],
    [{"url": "http://cache-test/URL"}],
    [{"url": "http://cache-test/URL/"}],
    # The ordering of the parameters in the URL string matters. They're not parsed at all, so a final '&' is also a cache breaker
    [{"url": "http://cache-test/TEST?a=1&b=2"}],
    [{"url": "http://cache-test/TEST?b=2&a=1"}],
    [{"url": "http://cache-test/TEST?a=1&b=2&"}],
    # `params` are also part of the URL key
    [{"url": "http://cache-test/params", "params": {}}],
    [{"url": "http://cache-test/params", "params": {"x": "a"}}],
    # `params` get appended to the URL, so these are equivalent. Using OrderedDict for pythons before 3.7.
    [
        {"url": "http://cache-test/params-test", "params": OrderedDict([("a", "1"), ("b", "2")])},
        {"url": "http://cache-test/params-test?", "params": OrderedDict([("a", "1"), ("b", "2")])},
        {"url": "http://cache-test/params-test?a=1", "params": OrderedDict([("b", "2")])},
        {"url": "http://cache-test/params-test?a=1&b=2", "params": OrderedDict()},
    ],
    # The presence, absence, or value of a header are all enough to bust the cache
    [{"url": "http://cache-test/header-test", "headers": {}}],
    [{"url": "http://cache-test/header-test", "headers": {"X-Test": "1"}}],
    [{"url": "http://cache-test/header-test", "headers": {"X-Test": "2"}}],
    # The `requests` library will normalise the case of headers before sending them off to the server, and so headers are
    # case-insensitive, and so all of these get the same cache key
    [
        {"url": "http://cache-test/header-test-2", "headers": {"X-Test": "1"}},
        {"url": "http://cache-test/header-test-2", "headers": {"x-test": "1"}},
        {"url": "http://cache-test/header-test-2", "headers": {"X-TEST": "1"}},
    ],
    # The order of the headers in the dict doesn't matter of course
    [
        {"url": "http://cache-test/header-order-test", "headers": {"X-Test-A": "a", "X-Test-B": "b"}},
        {"url": "http://cache-test/header-order-test", "headers": {"X-Test-B": "b", "X-Test-A": "a"}},
    ],
    # Setting a cookie via `cookies` is the same as setting it manually in a header
    [
        {"url": "http://cache-test/cookie-test", "cookies": {"a": "1"}, "headers": {}},
        {"url": "http://cache-test/cookie-test", "cookies": {}, "headers": {"Cookie": "a=1"}},
    ],
    # Setting `data` as a dict is equivalent to setting it manually as bytes
    [
        {"url": "http://cache-test/data-test", "data": {"a": "1"}},
        {
            "url": "http://cache-test/data-test",
            "data": b"a=1",
            "headers": {"Content-Type": "application/x-www-form-urlencoded", "Content-Length": "3"},
        },
    ],
    # Setting `json` is equivalent to manually building the request
    [
        {"url": "http://cache-test/json-test", "json": {"a": "1"}},
        {
            "url": "http://cache-test/json-test",
            "data": b'{"a": "1"}',
            "headers": {"Content-Type": "application/json"},
        },
    ],
    # requests can be specified as `Request` objects
    [
        {"url": "http://cache-test/request-objects", "method": "GET"},
        {"url": "http://cache-test/request-objects", "method": "GET", "params": {}},
        {"url": Request(url="http://cache-test/request-objects", method="GET")},
        {"url": Request(url="http://cache-test/request-objects", method="GET", params={})},
    ],
    [
        {"url": "http://cache-test/request-objects", "method": "POST", "data": {}},
        {"url": Request(url="http://cache-test/request-objects", method="POST", data={})},
    ],
    [
        {
            "url": "http://cache-test/request-objects",
            "method": "POST",
            "data": {"a": "1"},
        },
        {
            "url": "http://cache-test/request-objects",
            "method": "POST",
            "data": "a=1",
            "headers": {"Content-Type": "application/x-www-form-urlencoded"},
        },
        {
            "url": Request(
                url="http://cache-test/request-objects",
                method="POST",
                data={"a": "1"},
            ),
        },
        {
            "url": Request(
                url="http://cache-test/request-objects",
                method="POST",
                data="a=1",
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            ),
        },
    ],
]


@pytest.mark.parametrize(
    "config_1, config_2",
    iter_nonequal_pairs(EQUIVALENCIES),
)
def test_unique_keys(client, config_1, config_2):
    key = CacheKey.compute(dummy_compiled_request(client, **config_1), Config())
    other_key = CacheKey.compute(dummy_compiled_request(client, **config_2), Config())
    assert key != other_key


@pytest.mark.parametrize(
    "config_1, config_2",
    iter_nonequal_pairs(EQUIVALENCIES),
)
def test_unique_requests(client, config_1, config_2):
    cache = client.cache
    creq_1 = dummy_compiled_request(client, **config_1)
    creq_2 = dummy_compiled_request(client, **config_2)
    cache.put(creq_1, LogEntry(creq_1), dummy_response(creq_1))
    assert cache.get(creq_2, LogEntry(creq_2)) is None


@pytest.mark.parametrize(
    "config_1, config_2",
    iter_equal_pairs(EQUIVALENCIES),
)
def test_equivalent_keys(client, config_1, config_2):
    key_1 = CacheKey.compute(dummy_compiled_request(client, **config_1), Config())
    key_2 = CacheKey.compute(dummy_compiled_request(client, **config_2), Config())
    print(dummy_compiled_request(client, **config_1))
    print(dummy_compiled_request(client, **config_2))
    assert key_1 == key_2, (config_1, config_2)


@pytest.mark.parametrize(
    "config_1, config_2",
    iter_equal_pairs(EQUIVALENCIES),
)
def test_equivalent_requests(client, config_1, config_2):
    cache = client.cache
    creq_1 = dummy_compiled_request(client, **config_1)
    creq_2 = dummy_compiled_request(client, **config_2)
    response = dummy_response(creq_1)
    assert cache.get(creq_2, LogEntry(creq_2)) is None  # else test is invalid
    cache.put(creq_1, LogEntry(creq_1), response)
    assert cache.get(creq_2, LogEntry(creq_2)) == replace(response, from_cache=True)


def test_cache_updates_log_entry_attributes(client):
    cache = client.cache
    creq = dummy_compiled_request(client)
    log = LogEntry(creq)
    assert log.cache_key_str is None
    assert log.cached is None
    cache.get(creq, log)
    assert log.cache_key_str is not None
    assert log.cached is False
    cache.put(creq, log, dummy_response(creq))
    cache.get(creq, log)
    assert log.cached is True


TEST_KEYS = [
    ("simple-string", "/simple-string", "simple-string"),
    ("space string", "/space%20string", "space%20string"),
    ("slash/string", "/slash/string", "slash/string"),
    ("/slash/string", "/slash/string", "slash/string"),
    ("slash/string/", "/slash/string", "slash/string"),
    (("item", "123"), "/item/123", "item/123"),
    (("slash", "/"), "/slash/%2F", "slash/%2F"),
]


@pytest.mark.parametrize("user_specified, expected_path, expected_unique_str", TEST_KEYS)
def test_cache_key_parsing(user_specified, expected_path, expected_unique_str):
    parsed = CacheKey.parse(user_specified)
    assert "".join(f"/{p}" for p in parsed.path_parts) == expected_path
    assert parsed.unique_str == expected_unique_str


@pytest.mark.parametrize("user_specified", [spec[0] for spec in TEST_KEYS])
def test_cache_key_parsing_from_path_parts(user_specified):
    parsed = CacheKey.parse(user_specified)
    assert CacheKey.from_path_parts(parsed.path_parts) == parsed


def test_user_specified_cache_key(client, server):
    counter = count()
    all_keys = ["one", "two", "three"]
    all_values = [
        client.get(f"{server}/unique-number", cache_key=key, params={"unique": str(next(counter))}).text for key in all_keys
    ]
    assert len(set(all_values)) == len(all_values)  # they're all different
    for key, expected in zip(all_keys, all_values):
        obtained = client.get(
            f"{server}/unique-number",
            cache_key=key,
            # the param is actually different, but the cache key isn't, so we should get the same value back
            params={"unique": str(next(counter))},
        ).text
        assert obtained == expected


def test_user_specified_cache_key_on_redirect(client, server):
    res = client.get(
        f"{server}/redirect/chain/1",
        cache_key="fixed",
    )
    assert res.text == "Landed"
    all_keys_in_cache = sorted(k.unique_str for k in client.cache.storage.iter_all_keys())
    assert all_keys_in_cache == ["fixed", "fixed.1", "fixed.2"]


@pytest.mark.parametrize(
    "key_1, key_2, should_match",
    [
        (key_1, key_2, group_1 is group_2)
        for group_1, group_2 in product(
            # Keys from the same line here should match each other, and keys from different lines shouldn't
            [
                (CacheKey(parts=("a", "b")), ("a", "b"), "a/b"),
                (CacheKey(parts=("A", "B")), ("A", "B"), "A/B"),
                (CacheKey(parts=("b", "c")), ("b", "c"), "b/c"),
            ],
            repeat=2,
        )
        for key_1 in group_1
        for key_2 in group_2
    ],
)
def test_different_ways_to_express_cache_keys(client, server, key_1, key_2, should_match):
    counter = count()
    response_1, response_2 = (
        client.get(f"{server}/unique-number", cache_key=key, params={"unique": str(next(counter))}).text for key in (key_1, key_2)
    )
    if should_match:
        assert response_1 == response_2
    else:
        assert response_1 != response_2


def test_cache_key_of_unknown_class(client):
    class MyRandoClass:
        pass

    with pytest.raises(TypeError):
        client.fetch(
            "http://whatever/",
            cache_key=MyRandoClass(),
        )


def test_headers_ignored_by_cache(client):
    key1 = CacheKey.compute(
        dummy_compiled_request(client, headers={"Accept": "pudding", "Reject": "asparagus"}),
        Config(),
    )
    key2 = CacheKey.compute(
        dummy_compiled_request(client, headers={"Accept": "pudding", "Reject": "broccoli"}),
        Config(),
    )
    assert key1 != key2

    key1 = CacheKey.compute(
        dummy_compiled_request(client, headers={"Accept": "pudding", "Reject": "asparagus"}),
        Config(headers_ignored_by_cache=["Reject"]),
    )
    key2 = CacheKey.compute(
        dummy_compiled_request(client, headers={"Accept": "pudding", "Reject": "broccoli"}),
        Config(headers_ignored_by_cache=["Reject"]),
    )
    assert key1 == key2



#!/usr/bin/env python3

"""
Functions for serialising request and response objects to binary blobs, which can then be stored in any storage.

The format used is designed to mimic the HTTP exchange that travels on the wire, i.e. the blob starts with an loose rendering of
the HTTP request head, then request body if present, then the response head, then body. This makes it convenient to inspect the
blobs for manual debugging.
"""

# standards
from codecs import getwriter
from io import BytesIO
import re
from typing import Callable, Optional, Tuple

# hublot
from ..datastructures import CompiledRequest, Headers, Response


EOL = "\r\n"
EOL_BYTES = EOL.encode("UTF-8")
EOL_LEN = len(EOL)


Writer = getwriter("UTF-8")  # pylint: disable=invalid-name


def compose_binary_blob(res: Response) -> bytes:
    output = BytesIO()
    write = Writer(output).write
    _compose_request_blob(res.request, output, write)
    _compose_response_blob(res, output, write)
    return output.getvalue()


def _compose_request_blob(
    creq: CompiledRequest,
    output: BytesIO,
    write: Callable[[str], None],
) -> None:
    write(f"{creq.method} {creq.url}{EOL}")
    for key, value in sorted(creq.headers.items()):
        write(f"{key}: {value}{EOL}")
    write(EOL)
    if "Content-Length" in creq.headers:
        body = b"" if creq.data is None else creq.data
        content_length = int(creq.headers["Content-Length"])
        if content_length != len(body):
            # Don't write it out because we won't be able to read it back
            raise Exception(f"body has {len(body)} bytes but Content-Length is {content_length}")
        output.write(body)
        write(EOL)
    else:
        assert creq.data is None, "compile_request should've set the Content-Length"
    write(EOL)


def _compose_response_blob(
    res: Response,
    output: BytesIO,
    write: Callable[[str], None],
) -> None:
    write(f"HTTP {res.status_code} {res.reason}{EOL}")
    for key, value in sorted(res.headers.items()):
        write(f"{key}: {value}{EOL}")
    write(EOL)
    if res.content is not None:
        output.write(res.content)


def parse_binary_blob(data: bytes) -> Response:
    pos = 0
    method, url, pos = _parse_line(data, pos, r"^(\w+) (.+)$")
    req_headers, req_body, pos = _parse_message(data, pos)
    creq = CompiledRequest(
        url=url,
        method=method,
        headers=req_headers,
        data=req_body,
        num_retries=0,
    )
    status_code, reason, pos = _parse_line(data, pos, r"^HTTP (\d+) (.*)$")
    res_headers, res_body, pos = _parse_message(data, pos, read_to_end=True)
    assert res_body is not None  # since we passed `read_to_end=True`
    return Response(
        request=creq,
        from_cache=True,
        history=[],  # set elsewhere
        status_code=int(status_code),
        reason=reason,
        headers=res_headers,
        content=res_body,
    )


def _parse_message(
    data: bytes,
    pos: int,
    read_to_end: bool = False,
) -> Tuple[Headers, Optional[bytes], int]:
    headers = Headers()
    while data[pos : pos + EOL_LEN] != EOL_BYTES:
        key, value, pos = _parse_line(data, pos, r"^([^:]+): (.*)$")
        headers[key] = value
    pos += EOL_LEN
    body: Optional[bytes] = None
    if read_to_end:
        body = data[pos:]
        pos = len(data)
    elif headers.get("Content-Length"):
        length = int(headers["Content-Length"])
        body = data[pos : pos + length]
        pos += length + (2 * EOL_LEN)
    else:
        pos += EOL_LEN
    return headers, body, pos


def _parse_line(data: bytes, pos: int, regex: str):
    eol_pos = data.find(EOL_BYTES, pos)
    line = data[pos:eol_pos].decode("UTF-8")
    match = re.search(regex, line)
    if not match:  # pragma: no cover
        raise ValueError(repr(line))
    return (*match.groups(), eol_pos + EOL_LEN)



#!/usr/bin/env python3

# standards
import logging
import re
import subprocess
from typing import Iterable, Optional, Tuple
from urllib.parse import urlparse

# hublot
from ..config import Config
from ..datastructures import CompiledRequest, ConnectionError, Headers, Response, HublotException
from .base import Engine
from .pycurl import RE_HEADER, RE_STATUS
from .register import register_engine


LOGGER = logging.getLogger(__name__)


class CurlCmdEngineError(HublotException):
    """
    These exceptions indicate problems that are outside of HTTP errors -- unable to invoke the `curl` command, or to parse its
    output. This would be a bug or a system setup problem.
    """


class CurlCmdEngine(Engine):

    id = "curlcmd"

    def __init__(self, curl_cmd: str = "curl") -> None:
        self.curl_cmd = curl_cmd

    def short_code(self) -> str:
        return "cc"

    def request(self, creq: CompiledRequest, config: Config) -> Response:
        curl = subprocess.run(
            list(self._compose_curl_command(creq, config)),
            input=creq.data,
            capture_output=True,
            check=False,
        )
        if curl.returncode != 0:  # pragma: no cover
            output = curl.stderr.decode("UTF-8")
            message_match = re.search(rf"curl: \({curl.returncode}\) (.+)", output)
            message = message_match[1] if message_match else None
            if curl.returncode == 6:
                raise ConnectionError(message or output)
            else:
                raise HublotException(output)

        headers_match = re.search(rb"\r?\n\r?\n", curl.stdout)
        if not headers_match:  # pragma: no cover
            raise Exception("Failed to find headers in curl output")
        headers_str = curl.stdout[: headers_match.start()].decode("ISO-8859-1")

        status_code, reason, headers_str = self._parse_status_line(headers_str)
        return Response(
            request=creq,
            from_cache=False,
            history=[],
            status_code=status_code,
            reason=reason,
            headers=self._parse_headers(headers_str),
            content=curl.stdout[headers_match.end() :],
        )

    def _compose_curl_command(self, creq: CompiledRequest, config: Config) -> Iterable[str]:
        yield from [
            self.curl_cmd,
            creq.url,
            "--request",
            creq.method,
            "--connect-timeout",
            str(config.timeout),
            "--compressed",
            "--include",
        ]
        for key, value in creq.headers.items():
            yield from ["-H", f"{key}: {value}"]
        if creq.data is not None:
            yield from ["--data-binary", "@-"]
        if config.proxies:
            scheme = urlparse(creq.url).scheme
            proxy = config.proxies.get(scheme)
            if proxy:
                yield from ["--proxy", proxy]
        if not config.verify:
            yield "--insecure"

    @staticmethod
    def _parse_status_line(headers_str: str) -> Tuple[int, Optional[str], str]:
        match = RE_STATUS.search(headers_str)
        if not match:  # pragma: no cover
            raise HublotException("Malformed headers")
        return int(match[1]), match[2], headers_str[match.end() :]

    @staticmethod
    def _parse_headers(headers_str: str) -> Headers:
        headers = Headers()
        for match in RE_HEADER.finditer(headers_str):
            headers.add(match[1], match[2])
        return headers


register_engine(CurlCmdEngine)



#!/usr/bin/env python3

# standards
from collections.abc import Container
from dataclasses import asdict, dataclass
from datetime import timedelta
from typing import Dict, Optional, Tuple

# hublot
from .version import HUBLOT_VERSION


# Headers in this set will not be taken into account when computing cache keys, so two requests with different values for these
# headers will still get the same cache key.
#
DEFAULT_HEADERS_IGNORED_BY_CACHE = frozenset(
    [
        # By excluding User-Agent from cache we can upgrade Hublot, or fake a more recent browser, without zapping the entire cache.
        "User-Agent",
    ]
)


@dataclass
class Config:
    allow_redirects: bool = True
    cookies_enabled: bool = True
    courtesy_sleep: Optional[timedelta] = timedelta(seconds=5)
    force_cache_stale: bool = False
    headers_ignored_by_cache: Container[str] = DEFAULT_HEADERS_IGNORED_BY_CACHE
    max_cache_age: Optional[timedelta] = None
    max_redirects: int = 10
    proxies: Optional[Dict[str, str]] = None
    raise_for_status: bool = True
    timeout: Optional[int] = 60
    user_agent: Optional[str] = f"hublot/{HUBLOT_VERSION}"
    verify: Optional[bool] = True

    def derive_using_kwargs(self, **kwargs: object) -> Tuple["Config", Dict[str, object]]:
        # NB this must always return a new instance, so that the caller can modify it without affecting the original
        config = Config(**{key: kwargs.pop(key, default) for key, default in asdict(self).items()})  # type: ignore
        return config, kwargs



#!/usr/bin/env python3

# standards
from datetime import timedelta
from pathlib import Path
from tempfile import TemporaryDirectory
from urllib.parse import urlencode

# 3rd parties
import pytest
import requests

# hublot
from hublot import Cache, HttpClient, HttpError, Request, TooManyRedirects
from hublot.cache.storage import DiskStorage


@pytest.mark.parametrize(
    "kwargs, expected_method",
    [
        ({}, "GET"),
        ({"data": None}, "GET"),
        ({"data": b""}, "POST"),
        ({"data": b"a=b"}, "POST"),
        ({"data": {"a": "b"}}, "POST"),
        ({"json": None}, "GET"),
        ({"json": ""}, "POST"),
        ({"json": {"a": "b"}}, "POST"),
    ],
)
def test_default_method(client, server, kwargs, expected_method) -> None:
    assert client.fetch(f"{server}/method-test", **kwargs).text == expected_method


def test_no_cache_by_default(server) -> None:
    client = HttpClient()
    one = client.get(f"{server}/unique-number").text
    two = client.get(f"{server}/unique-number").text
    assert one != two  # not cached


def test_null_cache(server) -> None:
    client = HttpClient(cache=None)
    one = client.get(f"{server}/unique-number").text
    two = client.get(f"{server}/unique-number").text
    assert one != two  # not cached


def test_cache_as_path(server) -> None:
    with TemporaryDirectory() as tmp:
        client = HttpClient(cache=Path(tmp))
        one = client.get(f"{server}/unique-number").text
        two = client.get(f"{server}/unique-number").text
    assert one == two  # cached


def test_cache_as_cache_object(server) -> None:
    with TemporaryDirectory() as tmp:
        client = HttpClient(cache=Cache(DiskStorage(Path(tmp))))
        one = client.get(f"{server}/unique-number").text
        two = client.get(f"{server}/unique-number").text
    assert one == two  # cached


def test_force_cache_stale(client, server) -> None:
    one = client.get(f"{server}/unique-number").text
    two = client.get(f"{server}/unique-number", force_cache_stale=True).text
    three = client.get(f"{server}/unique-number").text
    assert one != two  # cache not read on 2nd call
    assert two == three  # but cache was written on 2nd call


def test_courtesy_sleep_by_default(mocked_courtesy_sleep, server) -> None:
    client = HttpClient()
    client.get(f"{server}/unique-number")
    client.get(f"{server}/unique-number")
    mocked_courtesy_sleep.assert_called_once()
    (delay,) = mocked_courtesy_sleep.call_args[0]
    assert delay > 1


def test_null_courtesy_sleep(mocked_courtesy_sleep, server) -> None:
    client = HttpClient(courtesy_sleep=None)
    client.get(f"{server}/unique-number")
    client.get(f"{server}/unique-number")
    mocked_courtesy_sleep.assert_not_called()


def test_custom_courtesy_sleep(mocked_courtesy_sleep, server) -> None:
    client = HttpClient(courtesy_sleep=timedelta(minutes=2))
    client.get(f"{server}/unique-number")
    client.get(f"{server}/unique-number")
    mocked_courtesy_sleep.assert_called_once()
    (delay,) = mocked_courtesy_sleep.call_args[0]
    assert delay == pytest.approx(120, 0.1)


def test_http_errors_are_raised(client, server) -> None:
    with pytest.raises(HttpError):
        client.get(f"{server}/fail-with-random-value")


def test_auto_raise_can_be_disabled(client, server) -> None:
    res = client.get(f"{server}/fail-with-random-value", raise_for_status=False)
    assert res.status_code == 500
    assert not res.ok


def test_redirect(client, server) -> None:
    res = client.get(f"{server}/redirect/chain/1")
    assert res.status_code == 200
    assert res.ok
    assert res.text == "Landed"


def test_no_redirect(client, server) -> None:
    res = client.get(f"{server}/redirect/chain/1", allow_redirects=False)
    assert res.status_code == 302
    assert res.text == "Bounce 1"


def test_redirect_response_bodies(cache, server) -> None:
    for _ in (1, 2):
        client = HttpClient(cache=cache)
        res = client.get(f"{server}/redirect/chain/1")
        assert res.status_code == 200
        assert res.text == "Landed"


def test_redirects_set_response_history(cache, server) -> None:
    for _ in (1, 2):
        client = HttpClient(cache=cache)
        res = client.get(f"{server}/redirect/chain/1")
        assert [r.text for r in res.history] == ["Bounce 1", "Bounce 2"]


def test_redirect_loop(client, server) -> None:
    # Make sure that the caching doesn't interfere with Requests' ability to detect redirect loops
    with pytest.raises(TooManyRedirects):
        client.fetch(f"{server}/redirect/loop")


@pytest.mark.parametrize(
    "request_method, body_kwarg",
    [
        ("GET", None),
        ("POST", "data"),
        ("POST", "json"),
        ("SLURP", "data"),
    ],
)
@pytest.mark.parametrize(
    "redirect_code",
    [301, 302, 303, 307, 308],
)
@pytest.mark.parametrize(
    "params_as_dict",
    [True, False],
)
def test_redirect_method(client, server, request_method, body_kwarg, redirect_code, params_as_dict) -> None:
    expected_method_for_redirected_request = request_method if redirect_code in (307, 308) else "GET"
    params = {"code": redirect_code, "something": "else"}
    res = client.request(
        method=request_method,
        url=f"{server}/redirect" + ("" if params_as_dict else f"?{urlencode(params)}"),
        params=params if params_as_dict else None,
        data="blabla" if request_method != "GET" and body_kwarg == "data" else None,
        json={"bla": "bla"} if request_method != "GET" and body_kwarg == "json" else None,
        headers={"Magic": "Mushroom"},
    )
    payload = res.json()
    assert payload["headers"]["Magic"] == "Mushroom"
    assert payload["method"] == expected_method_for_redirected_request
    assert payload["args"] == {"something": "else"}
    assert payload["data"] == (
        "" if expected_method_for_redirected_request == "GET" else "blabla" if body_kwarg == "data" else '{"bla":"bla"}'
    )


def test_client_preserves_casing_of_percent_escapes_in_path(client, server) -> None:
    ref_upper = client.get(f"{server}/bicam%C3%A9ral").text
    assert ref_upper.startswith("upper")
    ref_lower = client.get(f"{server}/bicam%c3%a9ral").text
    assert ref_lower.startswith("lower")
    assert client.get(f"{server}/bicam%C3%A9ral").text == ref_upper
    assert client.get(f"{server}/bicam%c3%a9ral").text == ref_lower


def test_client_preserves_casing_of_percent_escapes_in_query(client, server) -> None:
    ref_upper = client.get(f"{server}/bicam%C3%A9ral?name=Zo%C3%A9").text
    assert ref_upper.startswith("upper")
    ref_lower = client.get(f"{server}/bicam%C3%A9ral?name=Zo%c3%a9").text
    assert ref_lower.startswith("lower")
    assert client.get(f"{server}/bicam%C3%A9ral?name=Zo%C3%A9").text == ref_upper
    assert client.get(f"{server}/bicam%C3%A9ral?name=Zo%c3%a9").text == ref_lower


def test_client_can_fetch_from_server_that_redirects_based_on_escape_code_case(client, server) -> None:
    url = f"{server}/redirig%C3%A9"
    with pytest.raises(requests.TooManyRedirects):
        requests.get(url, timeout=60)  # doesn't work with `requests`, boo
    assert client.get(url).text == "lower"  # Hublot can get around it though, hurray


def test_get_method(client, server) -> None:
    assert client.get(url=f"{server}/method-test").text == "GET"


def test_post_method(client, server) -> None:
    assert client.post(url=f"{server}/method-test").text == "POST"


def test_request_method(client, server) -> None:
    assert client.request(url=f"{server}/method-test", method="PUT").text == "PUT"


def test_setting_both_data_and_json(client) -> None:
    # `Request` objects are mutable, so we can conceivably allow having both fields set, as long as it's fixed before the request
    # is sent. So this raises no exception.
    req = Request(
        url="http://hublot.test/",
        data={"x": "1"},
        json={"x": "1"},
    )
    # This, however, raises an exception
    with pytest.raises(TypeError):
        client.fetch(req)



#!/usr/bin/env python3

# 3rd parties
import pytest

# hublot
from hublot import HttpClient, HttpError


def test_http_error_is_raised(server):
    client = HttpClient()
    with pytest.raises(HttpError):
        client.get(f"{server}/fail-with-random-value")



#!/usr/bin/env python3

# standards
from pathlib import Path
import re

# 3rd parties
import setuptools


def get_version() -> str:
    version_file = Path(__file__).parent / "hublot" / "version.py"
    version_match = re.search(
        r"HUBLOT_VERSION = \"(.+)\"",
        version_file.read_text("UTF-8"),
    )
    if not version_match:
        raise ValueError("Couldn't parse version.py")
    return version_match.group(1)


setuptools.setup(
    name="hublot",
    version=get_version(),
    description="A thin wrapper around Requests that adds caching and throttling",
    url="https://github.com/saintamh/hublot",
    author="Hervé Saint-Amand",
    packages=setuptools.find_packages(exclude=["test"]),
    package_data={"hublot": ["py.typed"]},
    install_requires=[
        "chardet>= 3.0.2,<5",  # version req copied from requests
        'dataclasses>=0.8,<1; python_version<"3.7"',
        "requests>=2.25,<3",
    ],
    extras_require={
        "pycurl": [
            "pycurl>=7,<8",
        ],
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    zip_safe=False,  # https://mypy.readthedocs.io/en/latest/installed_packages.html#creating-pep-561-compatible-packages
)



#!/usr/bin/env python3

# standards
from contextlib import contextmanager
from http.cookiejar import DefaultCookiePolicy
import re

# 3rd parties
import requests
import urllib3.util.url

# hublot
from ..config import Config
from ..datastructures import CompiledRequest, ConnectionError, Headers, HublotException, Response
from .base import Engine
from .register import register_engine


class RequestsEngine(Engine):

    id = "requests"

    def __init__(self) -> None:
        self.session = requests.Session()
        # let hublot.HttpClient manage cookies
        self.session.cookies.set_policy(DefaultCookiePolicy(allowed_domains=[]))

    def short_code(self) -> str:
        return "rq"

    def request(self, creq: CompiledRequest, config: Config) -> Response:
        try:
            with patched_encode_invalid_chars():
                rres = self.session.request(
                    url=creq.url,
                    method=creq.method,
                    headers=creq.headers,  # type: ignore  # it fits in a duck-typed way
                    data=creq.data,
                    verify=config.verify,
                    allow_redirects=False,
                    timeout=config.timeout,
                )
                return Response(
                    request=creq,
                    from_cache=False,
                    history=[],
                    status_code=rres.status_code,
                    reason=rres.reason,
                    headers=Headers(rres.raw._fp.headers),  # pylint: disable=protected-access
                    content=rres.content,
                )
        except requests.exceptions.ConnectionError as error:
            raise ConnectionError(error) from error
        except requests.exceptions.RequestException as error:
            raise HublotException(error) from error


_encode_invalid_chars = urllib3.util.url._encode_invalid_chars  # type: ignore  # pylint: disable=protected-access


def _encode_invalid_chars_preserve_case(component: str, *rest, **kwargs) -> str:
    new = _encode_invalid_chars(component, *rest, **kwargs)
    if new != component:
        # `urllib3`, which is used internally by `requests`, changes the casing of %xx escapes, always forcing them to uppercase.
        # This means that if a server is set up such that when you ask for a URL with uppercase escapes it redirects you to the
        # same URL with lowercase escapes, then `requests` is unable to fetch that page -- the URL in the redirect response
        # 'location' will be re-uppercased, leading to a redirect loop. Here we work around that by chaging it back to lowercase.
        #
        # You could argue that the HTTP spec says %-escapes are case-insensitive, so it's the server described above that's
        # buggy, but browsers have no problem fetching such a page, so I'd argue it's a `requests` bug.
        #
        new = re.sub(r"%[0-9a-fA-F]{2}", lambda m: m.group().lower(), new)
    return new


@contextmanager
def patched_encode_invalid_chars():
    # pylint: disable=protected-access
    original = urllib3.util.url._encode_invalid_chars
    urllib3.util.url._encode_invalid_chars = _encode_invalid_chars_preserve_case
    try:
        yield
    finally:
        urllib3.util.url._encode_invalid_chars = original


register_engine(RequestsEngine)



#!/usr/bin/env python3

# standards
from dataclasses import dataclass, replace
import json
from typing import (
    # We use the title-cased Dict, List and Tuple for backwards compat with pythons <3.9
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

# 3rd parties
import chardet
import requests
from requests.cookies import MockRequest, MockResponse, RequestsCookieJar
from requests.utils import get_encoding_from_headers


JsonValue = Union[
    None,
    str,
    int,
    float,
    bool,
    List["JsonValue"],
    Tuple["JsonValue"],
    Dict[str, "JsonValue"],
]


class Headers:
    """
    A headers dict that uses case-insensitive keys and allows multiple values per key (for e.g. repeated "Set-Cookie" headers).

    Note that we deliberately don't inherit from `abc.Mapping` or similar because the interface isn't _quite_ that of a dict,
    because some methods return strings, and some return lists of strings.
    """

    _dict: Dict[str, List[Tuple[str, str]]]

    def __init__(self, base: Optional[Union["Headers", Dict[str, str]]] = None) -> None:
        self._dict = {}
        if base:
            for key, value in base.items():
                self.add(key, value)

    def __bool__(self) -> bool:
        return bool(self._dict)

    def __contains__(self, key: str) -> bool:
        return key.title() in self._dict

    def __getitem__(self, key: str) -> str:
        """
        Return the headers with the given key, as a single string. This is for convenience -- in 99.99% of cases users expect a
        single value, and don't want to deal with a list that will only have one element in it. Use `get_all` if you want all
        values.
        """
        value = self.get(key)
        if value is None:
            raise KeyError(key)
        return value

    def get(self, key: str, default: Optional[str] = None) -> Optional[str]:
        value_list = self.get_all(key)
        if not value_list:
            return default
        return "; ".join(value_list)

    def get_all(self, key: str, default: Sequence[str] = ()) -> Sequence[str]:
        value_list = self._dict.get(key.title())
        if value_list is None:
            return default
        return [value for _raw_key_unused, value in value_list]

    def add(self, key: str, value: str) -> None:
        self._dict.setdefault(key.title(), []).append((key, value))

    __setitem__ = add

    def setdefault(self, key: str, value: str) -> str:
        existing = self.get(key)
        if existing is not None:
            return existing
        else:
            self.add(key, value)
            return value

    def keys(self) -> Iterator[str]:
        """
        Yields a sequence of all header keys. Case-insensitive duplicates are removed.
        """
        for value_list in self._dict.values():
            yield value_list[0][0]

    __iter__ = keys

    def items(self, normalise_keys: bool = False) -> Iterator[Tuple[str, str]]:
        """
        Yield a sequence of all (key, value) header pairs. Note that unlike a proper dict, the sequence may include duplicated
        keys.

        Note that the `requests` library always normalises the casing of headers as they are sent out. For this reason our cache
        compares headers in a case-insensitive manner.
        """
        for normalised_key, value_list in self._dict.items():
            for raw_key, value in value_list:
                yield (normalised_key if normalise_keys else raw_key, value)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Headers):
            return False
        return sorted(self.items(normalise_keys=True)) == sorted(other.items(normalise_keys=True))

    def __repr__(self) -> str:
        return "Headers({%s})" % ", ".join(f"{key!r}: {value!r}" for key, value in self.items())


@dataclass
class Request:
    """
    An HTTP request, as formulated by the user code. This class is deliberately modelled on the requests library's eponymous
    class, to make easy for clients to switch from that library to this. Like that class, this one provides flexibility and
    convenience by allowing multiple redundant ways of specifying the same request -- for instance you can use `json` or `data`.

    This does not currently support streaming requests. They could be implemented, as long as the CompiledRequest has a way of
    representing them that is universal across engines, and as long as the cache handles them correctly (presumably by mandating
    that streamed requests must have a manually specified cache key).
    """

    url: str
    method: Optional[str] = None
    headers: Optional[Dict[str, str]] = None
    params: Optional[Dict[str, str]] = None
    data: Optional[Union[bytes, str, Dict[str, object]]] = None
    json: Optional[JsonValue] = None
    auth: Optional[Tuple[str, str]] = None
    cookies: Optional[Dict[str, str]] = None

    def replace(self, **kwargs) -> "Request":
        return replace(self, **kwargs)

    @classmethod
    def build(cls, raw: object, **kwargs) -> "Request":
        if isinstance(raw, str):
            return Request(url=raw, **kwargs)
        if isinstance(raw, Request):
            return raw.replace(**kwargs)
        if isinstance(raw, requests.Request):
            if raw.auth is not None and not isinstance(raw.auth, tuple):
                raise NotImplementedError(type(raw.auth))  # need to convert to a tuple
            if raw.cookies is not None and not isinstance(raw.cookies, dict):
                raise NotImplementedError(type(raw.cookies))  # need to convert to a dict
            merged_kwargs = dict(
                url=raw.url,
                method=raw.method,
                headers=raw.headers,
                params=raw.params,
                data=raw.data,
                json=raw.json,
                auth=raw.auth,
                cookies=raw.cookies,
            )
            merged_kwargs.update(kwargs)
            return cls(**merged_kwargs)
        raise ValueError(f"Don't know how to make a Request from {raw!r}")


Requestable = Union[str, Request, requests.Request]


@dataclass
class CompiledRequest:
    """
    Similarly to the requests library's Request/PreparedRequest duo, this class serves as a companion to the above Request
    class. This class is considered private to Hublot; instances are built internally, and client code is not expected to handle
    them. It represents the HTTP request, ready to be sent to the server (or, more exactly in the case of this library, ready to be
    submitted to the client engine.

    This ends up re-implementing the core of the requests library's CompiledRequest class, but it was needed so that we know
    we're submitting the exact same request to all engines.
    """

    url: str
    method: str
    headers: Headers
    data: Optional[bytes]
    num_retries: int


@dataclass
class Response:
    """
    Public class for HTTP response. As with `Request`, the interface is deliberately copied from the corresponding class in the
    requests library, to ease transition.
    """

    request: CompiledRequest
    from_cache: bool
    history: List["Response"]

    status_code: int
    reason: Optional[str]
    headers: Headers
    content: bytes

    @property
    def url(self) -> str:
        return self.request.url

    @property
    def encoding(self) -> Optional[str]:
        """The declared character encoding of the response, if any."""
        return get_encoding_from_headers(self.headers)

    @property
    def text(self) -> str:
        encoding = self.encoding
        if encoding is None:
            encoding = chardet.detect(self.content)["encoding"]
        if encoding is None:
            raise CharsetDetectionFailure()
        return self.content.decode(encoding)

    # NB `json` could have `JsonValue` as its return type, but then it means all client code needs to perform type checks on the
    # returned value (e.g. assert it's a dict etc). Most clients know what the data is and just want to get on with it. So we make
    # it Any. Clients can still do manual type checks if they want.
    def json(self, **kwargs) -> Any:
        return json.loads(self.text, **kwargs)

    def raise_for_status(self) -> None:
        if 400 <= self.status_code < 600:
            raise HttpError(
                f"{self.status_code} Client Error: {self.reason} for url: {self.url}",
                response=self,
            )

    @property
    def ok(self) -> bool:
        try:
            self.raise_for_status()
        except HttpError:
            return False
        return True

    @property
    def is_redirect(self) -> bool:
        return "Location" in self.headers and self.status_code in {301, 302, 303, 307, 308}

    @property
    def cookies(self) -> RequestsCookieJar:
        cookies = RequestsCookieJar()
        get_cookies_from_response(cookies, self)
        return cookies


def get_cookies_from_response(cookies: RequestsCookieJar, res: Response) -> None:
    """
    Copy cookies from the given `Response` to our cookie jar.
    """
    # This code copied from `requests.cookies.extract_cookies_to_jar`. The code is meant for requests's own data structures,
    # but ours are compatible enough to fit.
    cookies.extract_cookies(MockResponse(res.headers), MockRequest(res.request))  # type: ignore


class HublotException(Exception):
    pass


ScraperError = HublotException  # for backwards compatibility with Hublot <2


class HttpError(HublotException):
    def __init__(self, message: str, response: Response) -> None:
        super().__init__(message)
        self.response = response


class TooManyRedirects(HublotException):
    pass


class CharsetDetectionFailure(HublotException):
    pass


class ConnectionError(HublotException):
    pass
